/*
 * Test the automatically-generated DES implementations.
 *
 * Key should be - 5e d9 20 4f ec e0 b9 67
 *
 * Written by Matthew Kwan - April 1997.
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/time.h>

static const unsigned long	prac_k[50] = {
#include "prac-k64.h"
};

		/* External functions */
extern unsigned long	desprac64 (const unsigned long *inputs);


/*
 * Return an odd parity bit for a number.
 */

static int
odd_parity (
	int	n
) {
	int	parity = 1;

	while (n != 0) {
	    if ((n & 1) != 0)
		parity ^= 1;
	    n >>= 1;
	}

	return (parity);
}


/*
 * The key has been found.
 * Turn the value into something readable, and print it out.
 */

static void
key_found (
	unsigned long	slice,
	unsigned long	*rawkeys
) {
	int		i, key_idx = 0;
	int		keybits[56];

	keybits[3] = ((slice & 0x0f0f0f0f0f0f0f0f) != 0) ? 1 : 0;
	keybits[11] = ((slice & 0x0000ffff0000ffff) != 0) ? 1 : 0;
	keybits[13] = ((slice & 0x3333333333333333) != 0) ? 1 : 0;
	keybits[26] = ((slice & 0x00ff00ff00ff00ff) != 0) ? 1 : 0;
	keybits[41] = ((slice & 0x5555555555555555) != 0) ? 1 : 0;
	keybits[47] = ((slice & 0x00000000ffffffff) != 0) ? 1 : 0;

	key_idx = 55;
	for (i=0; i<50; i++) {
	    if (key_idx == 3 || key_idx == 11 || key_idx == 13 || key_idx == 26
					|| key_idx == 41 || key_idx == 47)
		key_idx--;

	    keybits[key_idx--] = (rawkeys[i] != 0) ? 1 : 0;
	}

	printf ("Key:");

	for (i=0; i<8; i++) {
	    int		j, kc = 0;

	    for (j=0; j<7; j++)
		if (keybits[49 - i*7 + j] != 0)
		    kc |= (1 << j);

	    kc = (kc << 1) | odd_parity (kc);

	    printf (" %02x", kc);
	}

	printf ("\n");
}


/*
 * Using the first 46 entries of the key as a counter, iterate
 * to the next value.
 */

static void
increment_key (
	unsigned long	*key
) {
	register unsigned long	*kp = &key[45];

	while (*kp != 0)
	    *kp-- = 0;

	*kp = ~0UL;
}


/*
 * Search 1024 DES keys.
 * Gray-code step through 16 calls to the 64-key evaluation function.
 */

static void
search1024 (
	unsigned long	*key
) {
	unsigned long	res;

	if ((res = desprac64 (key)) != 0)
	    key_found (res, key);
	key[49] = ~0UL;
	if ((res = desprac64 (key)) != 0)
	    key_found (res, key);
	key[48] = ~0UL;
	if ((res = desprac64 (key)) != 0)
	    key_found (res, key);
	key[49] = 0;
	if ((res = desprac64 (key)) != 0)
	    key_found (res, key);
	key[47] = ~0UL;
	if ((res = desprac64 (key)) != 0)
	    key_found (res, key);
	key[48] = 0;
	if ((res = desprac64 (key)) != 0)
	    key_found (res, key);
	key[49] = ~0UL;
	if ((res = desprac64 (key)) != 0)
	    key_found (res, key);
	key[48] = ~0UL;
	if ((res = desprac64 (key)) != 0)
	    key_found (res, key);
	key[46] = ~0UL;
	if ((res = desprac64 (key)) != 0)
	    key_found (res, key);
	key[48] = 0;
	if ((res = desprac64 (key)) != 0)
	    key_found (res, key);
	key[49] = 0;
	if ((res = desprac64 (key)) != 0)
	    key_found (res, key);
	key[48] = ~0UL;
	if ((res = desprac64 (key)) != 0)
	    key_found (res, key);
	key[47] = 0;
	if ((res = desprac64 (key)) != 0)
	    key_found (res, key);
	key[49] = ~0UL;
	if ((res = desprac64 (key)) != 0)
	    key_found (res, key);
	key[48] = 0;
	if ((res = desprac64 (key)) != 0)
	    key_found (res, key);
	key[49] = 0;
	if ((res = desprac64 (key)) != 0)
	    key_found (res, key);
	key[46] = 0;
}

/*
 * Test the DES key evaluation routine for speed.
 */

static void
test_speed (void)
{
	int		i, t;
	double		td;
	unsigned long	key[50];
	struct timeval	start_tv, end_tv;
	const int	n = 4096;

	for (i=0; i<50; i++)
	    key[i] = 0;

		/* Do a dummy run to get the function loaded into memory */
	search1024 (key);

		/* Begin the actual run */
	gettimeofday (&start_tv, NULL);

	for (i=0; i<n; i++) {
	    search1024 (key);
	    increment_key (key);
	}

	gettimeofday (&end_tv, NULL);

	t = (end_tv.tv_sec - start_tv.tv_sec) * 1000000
				+ (end_tv.tv_usec - start_tv.tv_usec);
	td = (double) t / 1000000.0;

	printf ("Searched %.1f keys per second\n", (double) n * 1024.0 / td);
}


/*
 * Iterate through all 2^56 DES keys.
 */

static void
keysearch (void)
{
	long int	i;
	unsigned long	key[50];

	for (i=0; i<50; i++)
	    key[i] = prac_k[i];
	for (i=40; i<50; i++)
	    key[i] = 0;

	for (i = 0; i < 0x400000000000; i++) {
	    search1024 (key);
	    increment_key (key);
	}
}


/*
 * Entry point.
 */

void
main (
	int	argc,
	char	*argv[]
) {
	if (argc > 1 && argv[1][0] == '-' && argv[1][1] == 'P')
	    test_speed ();
	else
	    keysearch ();

	exit (0);
}
